{ Unit:            interpre.pas
  Autor:           Sascha Nemecek
  Matrikelnr.:     9825815
  Erstellt am:     09.04.1999
  Letzte nderung: 21.04.1999

  Kurzbeschreibung:

  Dies ist ein Interpreter fur eine einfache Programmiersprache, der
  die Berechnungen mit globalen Integer-Variablen und globalen,
  parameterlosen Prozeduren erlaubt.
  Variablen- und Prozedurnamen bestehen nur aus Buchstaben. Alle Befehle und
  Variablennamen werden in Grossbuchstaben konvertiert.

  Der Interpreter kann folgende Sequenzen von Kommandos  verarbeiten:

  Command:
    'LET ' Id ':=' Expression |
    'WRITE ' Expression |
    'CALL ' Id |
    'PROCEDURE ' Id ':' |
    'END'

  Jede Zeile enthaelt genau ein Command. Whitespaces (Leerzeichen, Tabulator)
  werden ueberlesen. Nachdem der Interpreter alle Zeilen eingelesen hat,
  wird das Programm interpretiert.
  Das Hauptprogramm hat sich immer nach etwaigen Prozeduren zu befinden.


  Erlauterung der Befehle:

  LET:
     Der Wert der Expression wird der Variablen zugewiesen.
     Die Verwendung von Prozedurnamen in LET ist nicht zulaessig.

  WRITE:
     Der Wert der dem Befehl folgenden Expression wird ausgegeben.

  PROCEDURE/END:
     Mit PROCEDURE wird eine globale, parameterlose Prozedur definiert,
     die den nach PROCEDURE angegebenen Namen traegt.
     Geschachtelte Prozedurdefinitionen sind nicht zulaessig
     (muss auch nicht behandelt werden). Alle Commands bis zu END
     gehoeren zur Definition der Prozedur.

  CALL:
     Mit CALL wird die angegebene Prozedur aufgerufen.
     Nach dem Ende der Abarbeitung der Prozedur wird die Abarbeitung in der
     Zeile nach dem CALL fortgesetzt. Prozeduraufrufe koennen auch
     innerhalb von Prozeduren erfolgen.


  Eine Eingabedatei kann so aussehen:

        PROCEDURE ZWEIMAL :
          LET RES:=(B*2)
        END

        PROCEDURE BERECHNE :
          LET B:=WERT
          CALL ZWEIMAL
          LET SUM:=RES
          LET B:=5
          CALL ZWEIMAL
          LET SUM:=(SUM+RES)
        END

        LET WERT=6
        CALL BERECHNE
        WRITE SUM

  Die Ausgabe dieses Programms ist "22".

  Datenstrukturen
  Vor der eigentlichen Abarbeitung des Programms wird ein Verzeichnis aller
  Variablen- und Prozedurnamen in Form einer Hashtabelle aufgebaut. Diese
  beinhaltet einen Zeiger, der auf das jeweilige Objekt zeigt. Um Variablen
  von Prozeduren unterscheiden zu knnen, werden die Keys jeweils um
  'V' (Variable) bzw. 'P' (Prozedur) erweitert.
  Variable werden in Form eines Parse Trees gespeichert.
  Prozeduren und das Hauptprogramm als lineare Liste.
  Damit waehrend der Interpretation eines Programms nach der Abarbeitung
  einer Prozedur der Programmablauf an der richtigen Stelle
  fortgesetzt werden kann, werden Ruecksprungreferenzen
  (Verweise auf die einem Aufruf folgende Anweisung) in Form eines Stacks
  verwaltet.
  }


PROGRAM interpre;

USES hashtbl, mystack, parse3, linlist2, token, magtools;


{ Definitionsteil. }
TYPE


{ Definiert den Aufbau eines Programmknotens. Um speichereffizient
  zu sein ist er als Variante aufgebaut. }
ProgArt = (LET, CALL, WRITE);
progKnotenPtr = ^progKnoten;
progKnoten = RECORD
   CASE art: ProgArt OF
      LET: (
         dest: STRING;
         source: pointer;);

      CALL: (
         callproc: STRING;);

      WRITE: (
         writevar: STRING;)
END;


interPtr = ^inter;
inter = OBJECT
   CONSTRUCTOR Init;
   DESTRUCTOR Done;

  PRIVATE
   { Hashtable mit Variablen und Prozeduren. }
   vp_index: hashptr;

   { Aktuelles Programm. }
   progPtr: ListePtr;

   PROCEDURE openFile;
   PROCEDURE runProgramme;
   PROCEDURE do_Let (a, b: STRING);
   PROCEDURE do_Procedure (s: STRING);
   PROCEDURE do_End;
   PROCEDURE do_Write (s: STRING);
   PROCEDURE do_Call (s: STRING);

END;



VAR
   runMe: interPtr;



{ Prozedure do_let.
  Verarbeitet die Interpreteranweisung "LET". Initialisiert einen
  Parse Tree um den Wert der Variablen zu speichern und trgt sie
  in den Prozeduren- & Variablenindex ein.
  @param a Der Variablenname.
  @param b Die Anweisung fr den Parsetree. }
PROCEDURE inter.do_Let (a, b: STRING);
VAR ptptr: parsePtr;
    hilf: progKnotenPtr;
    i: INTEGER;
BEGIN
   { Check ob Ausdrcke vorhanden. }
   IF (a = '') OR (b = '') THEN BEGIN
      writeln ('Fehler in folgendem Ausdruck: LET ', a, ':=', b);
      writeln ('Grund: Links- bzw. Rechtswert fehlt.');
      writeln ('PROGRAMMABBRUCH!');
      HALT;
   END;

   { Wenn das Hauptprogramm noch nicht initialisiert wurde. }
   IF progPtr = NIL THEN
      new (progPtr, Init);

   { Variable initialisieren und in Index eintragen. }
   IF vp_index^.find (a) = NIL THEN BEGIN

      { Variablensyntax berprfen. }
      FOR i := 1 TO length (a) DO
         IF isAlpha (a[i]) = FALSE THEN BEGIN
            writeln ('Fehler in folgendem Ausdruck: ', a, ':=', b);
            writeln ('Grund: Variable drfen nur aus Buchstaben bestehen.');
            writeln ('PROGRAMMABBRUCH!');
            HALT;
         END;

      new (ptptr, init (vp_index));
{      ptptr^.str2parse ('0');}
      vp_index^.Insert ('V' + a, ptptr);
   END;

   { Programmstruktur um Programmknoten erweitern. }
   new (hilf);
   hilf^.art := LET;
   hilf^.dest := a;

   { Expression Syntax prfen. }
   FOR i := 1 TO length (b) DO
      IF (isAlpha (b[i]) = FALSE) AND (isNum (b[i]) = FALSE) AND
         (b[i] <> '(') AND (b[i] <> ')') AND (b[i] <> '*') AND
         (b[i] <> '+') AND (b[i] <> '-') AND (b[i] <> '/') THEN BEGIN

         writeln ('Fehler in folgendem Ausdruck: ', a, ':=', b);
         writeln ('Grund: Ungltiges Zeichen in der Anweisung.');
         writeln ('PROGRAMMABBRUCH!');
         HALT;
      END;

   { Eintragen der Expression. }
   new (ptptr, init (vp_index));
   ptptr^.str2parse (b);
   hilf^.source := ptptr;
   progPtr^.addElement (hilf);
END;



{ Prozedure do_Procedure.
  Verarbeitet die Interpreteranweisung "PROCEDURE". Initialisiert
  eine Lineare Liste und trgt sie in den Prozeduren- &
  Variablenindex ein.
  @param s Der Prozedurname. }
PROCEDURE inter.do_Procedure (s: STRING);
VAR i: integer;
BEGIN
   { Wenn noch keine andere Prozedur bearbeitet wird. }
   IF progPtr = NIL THEN BEGIN
      new (progPtr, Init);
   END ELSE BEGIN
      writeln ('Fehler: In einer Prozedur kann keine weiter definiert werden');
      writeln ('        und das Hauptprogramm muss nach den Prozeduren folgen.');
      writeln ('PROGRAMMABBRUCH!');
      HALT;
   END;

   { Prozedursyntax berprfen. }
   FOR i := 1 TO length (s) DO
      IF isAlpha (s[i]) = FALSE THEN BEGIN
         writeln ('Fehler in folgendem Ausdruck: PROCEDURE ', s, ' :');
         writeln ('Grund: Prozeduren drfen nur aus Buchstaben bestehen.');
         writeln ('PROGRAMMABBRUCH!');
         HALT;
      END;

   { Prozedur initialisieren und in Index eintragen. }
   IF vp_index^.find (s) = NIL THEN BEGIN
      new (progPtr, init);
      vp_index^.Insert ('P' + s, progPtr);
   END;
END;



{ Prozedure do_End.
  Verarbeitet die Interpreteranweisung "END". Beendet die Prozedur. }
PROCEDURE inter.do_End;
BEGIN
   IF progPtr <> NIL THEN
      progPtr := NIL
   ELSE BEGIN
      writeln ('Fehler: Ungltiger Prozedurabbruch.');
      writeln ('PROGRAMMABBRUCH!');
      HALT;
   END;
END;


{ Prozedure do_Write.
  Verarbeitet die Interpreteranweisung "WRITE".
  @param s Die auszugebene Variable. }
PROCEDURE inter.do_Write (s: STRING);
VAR hilf: progKnotenPtr;
BEGIN
   { berprfung ob Variable angegeben wurde. }
   IF s = '' THEN BEGIN
      writeln ('Fehler: Keine Variable zur Ausgabe angegeben.');
      writeln ('PROGRAMMABBRUCH!');
      HALT;
   END;

   { Wenn das Hauptprogramm noch nicht initialisiert wurde. }
   IF progPtr = NIL THEN
      new (progPtr);

   new (hilf);
   hilf^.art := WRITE;
   hilf^.writevar := s;

   progPtr^.addElement (hilf);
END;



{ Prozedure do_Call.
  Verarbeitet die Interpreteranweisung "CALL". Initialisiert einen
  neue Programmliste.
  @param s Der Prozedurname. }
PROCEDURE inter.do_Call (s: STRING);
VAR hilf: progKnotenPtr;
    i: integer;
BEGIN
   { berprfung ob Prozedur angegeben wurde. }
   IF s = '' THEN BEGIN
      writeln ('Fehler: Keine Prozedur zum Aufruf angegeben.');
      writeln ('PROGRAMMABBRUCH!');
      HALT;
   END;

   { Wenn das Hauptprogramm noch nicht initialisiert wurde. }
   IF progPtr = NIL THEN
      new (progPtr);

   { Prozedursyntax berprfen. }
   FOR i := 1 TO length (s) DO
      IF isAlpha (s[i]) = FALSE THEN BEGIN
         writeln ('Fehler in folgendem Ausdruck: PROCEDURE ', s, ' :');
         writeln ('Grund: Prozeduren drfen nur aus Buchstaben bestehen.');
         writeln ('PROGRAMMABBRUCH!');
         HALT;
      END;

   { berprfen ob rekursiver Aufruf. }
   IF vp_index^.find ('P' + s) = progPtr THEN BEGIN
      writeln ('Fehler: Rekursive Aufrufe sind nicht erlaubt.');
      writeln ('PROGRAMMABBRUCH!');
      HALT;
   END;

   new (hilf);
   hilf^.art := CALL;
   hilf^.callproc := s;

   progPtr^.addElement (hilf);
END;


PROCEDURE inter.openFile;
VAR eindat: text;
    zeile, tok, help: STRING;
    ch: char;

BEGIN
   { berprfung der Parameteranzahl. }
   IF paramcount <> 1 THEN BEGIN
      writeln ('Ungltige Parameteranzahl!');
      writeln ('Syntax: INTERPRE <QUELLDATEI>');
      HALT;
   END;

   { Versucht die angegebene Datei zu ffnen }
   {$I-}
   assign (eindat, paramstr (1));
   reset (eindat);
   IF ioresult<>0 THEN BEGIN
      writeln ('Fehler beim ffnen der Datei !');
      HALT;
   END; {$I+}

   new (vp_index, init);
   progPtr := NIL;

   { Datei in Datenstruktur umwandeln. }
   WHILE NOT eof (eindat) DO BEGIN
      readln (eindat, zeile);

      { Zeile in Grossbuchstaben umwandeln. }
      strtoupper (zeile);

      { Suche nach Whitespaces. }
      tokenize (zeile, tok, ' ' + #9);
      WHILE length (tok) > 0 DO BEGIN

         { Begin einer Prozedur. }
         IF tok = 'PROCEDURE' THEN BEGIN
            tokenize ('', tok, '');
            help := tok;
            tokenize ('', tok, '');

            { berprfen der Syntax. }
            IF (help = '') THEN BEGIN
               writeln ('Fehler in folgendem Ausdruck: PROCEDURE ', help, ' ', tok);
               writeln ('Grund: Der Prozedurbezeichner fehlt.');
               writeln ('PROGRAMMABBRUCH!');
               HALT;
            END;

            { berprfen der Syntax. }
            IF tok <> ':' THEN BEGIN
               writeln ('Falsche Syntax bei Prozedurdefinition PROCEDURE ', help, ' - (: fehlt)!');
               writeln ('PROGRAMMABBRUCH!');
               HALT;
            END;

            do_Procedure (help);
            tok[0]:=#0;
         END ELSE


         { Ende einer Prozedur. }
         IF tok = 'END' THEN BEGIN
            do_End;
            tok[0]:=#0;
         END ELSE


         { Aufruf einer Prozedur. }
         IF tok = 'CALL' THEN BEGIN
            tokenize ('', tok, '');
            do_Call (tok);
            tok[0]:=#0;
         END ELSE


         { Zuweisen einer Variablen. }
         IF tok = 'LET' THEN BEGIN
            tokenize ('', tok, ' :' + #9);
            help := tok;
            tokenize ('', tok, '=');

            IF tok <> ':' THEN BEGIN
               writeln ('Fehler: LET ', help, tok, ' - ungltige Syntax.');
               writeln ('PROGRAMMABBRUCH!');
               HALT;
            END;

            tokenize ('', tok, ' ' + #9);
            strincpy (tok, tok, 2, length (tok) - 2);

            do_Let (help, tok);
            tok[0]:=#0;
         END ELSE


         { Variable ausgeben. }
         IF tok = 'WRITE' THEN BEGIN
            tokenize ('', tok, '');
            do_Write (tok);
            tok[0]:=#0;
         END ELSE

         { Sonstige Mglichkeiten (Leerzeilen, Falscheingaben). }
         IF tok <> #0 THEN BEGIN
            writeln ('Fehler: Unbekannter Befehl ', tok);
            writeln ('PROGRAMMABBRUCH!');
            HALT;
         END ELSE
            tok [0]:=#0;

      END;
   END;

   { Eingabedatei schliessen. }
   close (eindat);
END;


PROCEDURE inter.runProgramme;
VAR i: integer;
    progStack: stackPtr;
    hilfsstr: STRING;
    hilfsprog: progKnotenPtr;
    hilfsvar: parsePtr;
BEGIN
   { Stack initialisieren. }
   new (progStack, init);

   { Hier beginnt der Ausfhrungsteil. }
   progPtr^.gotoStart;

   WHILE TRUE DO BEGIN

      { Aktuelle Anweisung auslesen. }
      hilfsprog := progPtr^.getData;

      { Anweisung verarbeiten. }
      CASE hilfsprog^.art OF

         { Wert einer Variblen setzen. }
         LET: BEGIN
            {writeln ('LET');}

            { Quellwert berechnen. }
            hilfsvar := hilfsprog^.source;

            IF hilfsvar^.solvetree (i) = FALSE THEN BEGIN
               writeln ('Fehler: LET ', hilfsprog^.dest, ' konnte nicht berechnet werden!');
               writeln ('PROGRAMMABBRUCH!');
               HALT;
            END;

            { Ziel bestimmen und Wert setzen. }
            hilfsvar := vp_index^.find ('V' + hilfsprog^.dest);
            str (i, hilfsstr);
            hilfsvar^.setValue (hilfsstr);

            { Zr nchsten Anweisung. }
            IF progPtr^.eol = FALSE THEN
               progPtr^.gotoNext;
         END;

         { Call ausfhren. }
         CALL: BEGIN
            {writeln ('CALL'); }

            progPtr^.gotoNext;
            progStack^.push (progPtr);
            progPtr := vp_index^.find ('P' + hilfsprog^.callproc);

            IF progPtr = NIL THEN BEGIN
               writeln ('Fehler in folgendem Ausdruck: CALL ', hilfsprog^.callproc);
               writeln ('Grund: Prozedure wurde nicht definiert.');
               writeln ('PROGRAMMABBRUCH!');
               HALT;
            END;

            progPtr^.gotoStart;
         END;

         { Write Anweisung ausfhren. }
         WRITE: BEGIN
            hilfsvar := vp_index^.find ('V' + hilfsprog^.writevar);

            IF hilfsvar^.solvetree (i) = FALSE THEN BEGIN
               writeln ('Fehler: ', hilfsprog^.dest, ' konnte nicht berechnet werden!');
               HALT;
            END;

            writeln (hilfsprog^.writevar, '=', i);

            { Zr nchsten Anweisung. }
            IF progPtr^.eol = FALSE THEN
               progPtr^.gotoNext;
         END;

      END;

      { berprfen ob Programm- oder Prozedurende. }
      IF progPtr^.eol = TRUE THEN
         IF progStack^.isEmpty = TRUE THEN
            BREAK
         ELSE
            progPtr := progStack^.pop;
   END;

   { Stack wieder frei geben. }
   dispose (progStack, done);
END;



{ Constructor.
  Veranlasst das einlesen der Datei und deren Auswertung.
  Danach werden die Anweisungen ausgefhrt. }
CONSTRUCTOR inter.Init;
BEGIN
   openFile;
   runProgramme;
END;




{ Destructor.
  Gibt den belegten Speicherplatz wieder frei. }
DESTRUCTOR inter.Done;
BEGIN
   { Speicher wieder frei geben. }
   dispose (vp_index, done);
   dispose (progPtr, done);
END;



BEGIN
   new (runMe, Init);
   dispose (runMe, Done);
END.