{ Unit:            parse3.pas
  Autor:           Sascha Nemecek
  Matrikelnr.:     9825815
  Erstellt am:     11.04.1999
  Letzte nderung: 21.04.1999

  Kurzbeschreibung:
  Ermglicht es einfache arithmetische Ausdrcke in einem Parse Baum
  darzustellen. Er ist allerdings auf Integerausdrcke beschrnkt.
  Es stehen folgende Methoden zu Verfgung:

  Public:
     str2parse:     Versucht den bergebenen String in einen Parse Tree
                    umzuwandeln.
     solveTree:     Berechnet den Wert des Baumes.
     setValue:      Initialisiert den Baum neu.

  Private:
     clearAll:      Lscht den gesamten Baum.
     getOperand:    Teilt eine Expression und liefert den Index des
                    obersten Operanden zurck.

  Damit der Parse Tree auch Variablen verarbeiten kann ist es notwendig
  bei der Initialisierung den Variablenindex in Form einer Hashtable
  zu bergeben.
  }


UNIT parse3;


INTERFACE

USES hashtbl;



TYPE

operand = (Variable, Konstante);
KnotenArt = (Zwischenknoten, Endknoten);
parseKnotenPtr = ^parseKnoten;
parsePtr = ^parse;


{ Definiert den Aufbau eines Parseknotens. Um speichereffizient
  zu sein ist er als Variante aufgebaut. }
parseKnoten = RECORD
   CASE art: knotenArt OF
      Zwischenknoten: (
         operation: char;
         links, rechts: parseKnotenPtr;);
      Endknoten: (
         CASE op: operand OF
            Variable: (
               varptr: STRING;);
            Konstante: (
               konst: integer;));
END;



{ Klasse: Parse Tree. }
parse = OBJECT
   CONSTRUCTOR Init (vi: hashPtr);
   FUNCTION str2parse (s: STRING) : pointer;
   FUNCTION solveTree (var value: integer) : boolean;
   FUNCTION setValue (s: STRING) : pointer;
   DESTRUCTOR Done;

  PRIVATE
   varindex: hashPtr;
   root: parseKnotenPtr;
   FUNCTION getOperand (s: STRING): integer;
   PROCEDURE clearAll;
END;



IMPLEMENTATION



{ Constructor Init.
  Setzt alle Anfangswerte. Und der bergebene Variablenindex wird
  bernommen.
  @param vi Der bergeben Variablenindex in Form einer Hashtabelle. }
CONSTRUCTOR parse.Init (vi: hashPtr);
BEGIN
   varindex := vi;
   root := NIL;
END;



{ Funktion str2parse.
  Wandelt einen String in einen Parse Tree um. Bei Verschachtelten
  Ausdrcken wird die Funktion solange rekursiv aufgerufen, bis
  s auf einen einfachen Ausdrck gebracht wurde.
  Kann der String nicht umgewandelt werden wird NIL zurckgeliefert.
  @param s Der zu Parsende String.
  @return Der Pointer auf des Erstellte Objekt im Parse Baum. }
FUNCTION parse.str2parse (s: STRING) : pointer;
VAR h: parseKnotenPtr;
    ptptr: parsePtr;
    code, i: Integer;
    s1, s2: STRING;
    ok: boolean;
BEGIN
   { Neuen Knoten erzeugen. }
   new (h);

   { Wenn noch kein Knoten vorhanden, diesen als Wurzel verwenden. }
   IF root = NIL THEN
      root := h;
   ok := FALSE;

   { Auswerten des Ausdruckes. }
   CASE s[1] OF

      { Bei Konstante diese eintragen. }
      '0'..'9':  BEGIN
         h^.art := Endknoten;
         h^.op := Konstante;
         val (s, h^.konst, code);
         IF code = 0 THEN
            ok := TRUE;
      END;

      { Bei Variable diese eintragen. }
      'A'..'Z':  BEGIN
         h^.art := Endknoten;
         h^.op := Variable;
         h^.varptr := s;
         IF length (h^.varptr) <> 0 THEN
            ok := TRUE;
      END;

      { Bei Klammerung: weiter zerlegen. }
      '(': BEGIN
         h^.art := Zwischenknoten;
         code := getOperand (s);

         { Wenn Operand gefunden, OK. }
         IF code <> 0 THEN BEGIN
            { Ersten Teilstring berechnen. }
            s1 [0] := #0;
            FOR i := 2 TO (code-1) DO
               s1 := s1 + s[i];

            { Zweiten Teilstring berechnen. }
            s2 [0] := #0;
            FOR i := (code+1) TO (length (s)-1) DO
               s2 := s2 + s[i];

            h^.operation := s[code];
            h^.links := str2parse (s1);
            h^.rechts := str2parse (s2);
            IF (h^.links <> NIL) AND (h^.rechts <> NIL) THEN
               ok := TRUE;
         END ELSE
            ok := FALSE;
      END;
   END;

   { Wenn kein Fehler aufgetreten ist. }
   IF ok = TRUE THEN
      str2parse := h
   ELSE BEGIN
      { Wenn keiner der Flle zutrifft oder ein Fehler auftrat. }
      dispose (h);
      root := NIL;
      str2parse := NIL;
   END;
END;



{ Funktion solveTree.
  Berechnet den Endwert des Baumes. Tritt ein Fehler auf, wird
  FALSE zurckgegeben.
  @param value Die Variable, die gesetzt werden soll.
  @return TRUE Wenn die Berechnung fehlerfrei war.
          FALSE Wenn ein Fehler aufgetreten ist. }
FUNCTION parse.solveTree (var value: integer) : boolean;
VAR h: parseKnotenPtr;
    ptptr: parsePtr;
    ok: boolean;
    v1, v2: integer;
BEGIN
   IF root = NIL THEN BEGIN
      solveTree := FALSE;
      EXIT;
   END;

   h := root;
   ok:= TRUE;

   { Bei einem Zwischenknoten. }
   IF root^.art = Zwischenknoten THEN BEGIN

      { Linken Ast berprfen. }
      IF ((root^.links)^.art) = Zwischenknoten THEN BEGIN
         root := root^.links;
         if solveTree (v1) = FALSE THEN
            ok := FALSE;
         root := h;
      END ELSE
         IF root^.links^.op = Konstante THEN
            v1 := root^.links^.konst
         ELSE BEGIN
            ptptr := varindex^.find ('V' + root^.links^.varptr);

            { Falls Variable nicht existiert: Abbruch. }
            IF ptptr = NIL THEN
               ok := FALSE
            ELSE IF ptptr^.solveTree (v1) = FALSE THEN BEGIN
               writeln ('Fehler: ', root^.links^.varptr, ' nicht definiert.');
               writeln ('PROGRAMMABBRUCH!');
               HALT;
            END;
         END;

      { Rechten Ast berprfen. }
      IF ((root^.rechts)^.art) = Zwischenknoten THEN BEGIN
         root := root^.rechts;
         if solveTree (v2) = FALSE THEN
            ok := FALSE;
         root := h;
      END ELSE
         IF root^.rechts^.op = Konstante THEN
            v2 := root^.rechts^.konst
         ELSE BEGIN
            ptptr := varindex^.find ('V' + root^.rechts^.varptr);

            { Falls Variable nicht existiert: Abbruch. }
            IF ptptr = NIL THEN
               ok := FALSE
            ELSE IF ptptr^.solveTree (v2) = FALSE THEN BEGIN
               writeln ('Fehler: ', root^.rechts^.varptr, ' nicht definiert.');
               writeln ('PROGRAMMABBRUCH!');
               HALT;
            END;
         END;


      { Ergebnis des Zwischenknotens Berechnen. }
      CASE root^.operation OF
         '+': value := v1 + v2;
         '-': value := v1 - v2;
         '*': value := v1 * v2;
         '/': value := v1 DIV v2;
      END;
   END ELSE
      IF root^.op = Konstante THEN
         value := root^.konst
      ELSE BEGIN
         ptptr := varindex^.find ('V' + root^.varptr);

         { Falls Variable nicht existiert: Abbruch. }
         IF ptptr = NIL THEN
            ok := FALSE
         ELSE IF ptptr^.solveTree (value) = FALSE THEN BEGIN
            writeln ('Fehler: ', root^.varptr, ' nicht definiert.');
            writeln ('PROGRAMMABBRUCH!');
            HALT;
         END;
      END;

   IF ok = TRUE THEN
      solveTree := TRUE
   ELSE
      solveTree := FALSE;
END;



{ Funktion setValue.
  Lscht den aktuellen Baum und initialisiert ihn neu. Dabei
  darf es sich aber nur um einen Baum handeln, in dem ein Varible
  gespeichert ist.
  @param s Bezeichner der Variable.
  @return Pointer auf den Eintrag der Variablen. }
FUNCTION parse.setValue (s: STRING): pointer;
VAR h: parseKnotenPtr;
    code, i: Integer;
    s1, s2: STRING;
    ok: boolean;
BEGIN
   IF root = NIL THEN BEGIN
      new (h);
      root := h;
   END ELSE
      h := root;

   ok := FALSE;

   { Konstante eintragen. }
   h^.art := Endknoten;
   h^.op := Konstante;
   val (s, h^.konst, code);
   IF code = 0 THEN
      ok := TRUE;

   { Wenn kein Fehler aufgetreten ist. }
   IF ok = TRUE THEN
      setValue := h
   ELSE BEGIN
      { Wenn keiner der Flle zutrifft oder ein Fehler auftrat. }
      dispose (h);
      root := NIL;
      setValue := NIL;
   END;
END;



{ Funktion getOperand.
  Sucht den obersten Operator in einem String (Infix Notation).
  Liefert den Index des Operators zurck. Bei Fehler wird 0 zurckgegeben.
  @param s Der Operator.
  @return Index des Operators.
  }
FUNCTION parse.getOperand (s: STRING): integer;
VAR i, d: integer;
BEGIN
   d := 0;

   { Durchscannen des Ausdrucks ohne Anfangs- und Endklammer. }
   FOR i := 2 TO (length (s)-1) DO BEGIN
      CASE s[i] OF
         '(': inc (d);
         ')': dec (d);
         '+','-','*','/': IF d = 0 THEN getOperand := i;
      END;
   END;
END;



{ Prozedur clearAll.
  Lscht den ganzen Baum. }
PROCEDURE parse.clearAll;
VAR h: parseKnotenPtr;
BEGIN
   IF root = NIL THEN
      EXIT;

   h := root;

   { Bei einem Zwischenknoten. }
   IF root^.art = Zwischenknoten THEN BEGIN

      { Linken Ast berprfen. }
      IF ((root^.links)^.art) = Zwischenknoten THEN BEGIN
         root := root^.links;
         clearAll;
         root := h;
      END ELSE
         dispose (root^.links);

      { Rechten Ast berprfen. }
      IF ((root^.rechts)^.art) = Zwischenknoten THEN BEGIN
         root := root^.rechts;
         clearAll;
         root := h;
      END ELSE
         dispose (root^.rechts)

   END ELSE
      dispose (root);
END;



{ DESTRUCTOR Done.
  Lscht alle Elemente des Parse Trees. }
DESTRUCTOR parse.Done;
BEGIN
   clearAll;
END;



BEGIN
END.